-- LOIH Cal -------------------------------------------------------------------
-------------------------------------------------------------------------------
local _G = getfenv(0)
local ADDON_NAME, ns = ... -- Addon name and private namespace
local L = ns.L -- Localization table


-- DEBUG ----------------------------------------------------------------------
-------------------------------------------------------------------------------
local DEBUGGING = ns.DEBUGGING
local Debug = ns.Debug
local print_r = ns.print_r


-- UPVALUES, DEFAULTS AND LOCAL SCOPES ----------------------------------------
-------------------------------------------------------------------------------
-- API Upvalues
local table, ipairs, type, tostring, pairs = _G.table, _G.ipairs, _G.type, _G.tostring, _G.pairs
local next, string, select, setmetatable, rawset = _G.next, _G.string, _G.select, _G.setmetatable, _G.rawset

-- Constant Upvalues
local HIGHLIGHT_FONT_COLOR_CODE, FONT_COLOR_CODE_CLOSE = _G.HIGHLIGHT_FONT_COLOR_CODE, _G.FONT_COLOR_CODE_CLOSE
local RAID_CLASS_COLORS, RED_FONT_COLOR_CODE = _G.RAID_CLASS_COLORS, _G.RED_FONT_COLOR_CODE
local ORANGE_FONT_COLOR_CODE, GREEN_FONT_COLOR_CODE = _G.ORANGE_FONT_COLOR_CODE, _G.GREEN_FONT_COLOR_CODE
local NORMAL_FONT_COLOR_CODE, ITEM_QUALITY_COLORS = _G.NORMAL_FONT_COLOR_CODE, _G.ITEM_QUALITY_COLORS
local GRAY_FONT_COLOR, BATTLENET_FONT_COLOR = _G.GRAY_FONT_COLOR, _G.BATTLENET_FONT_COLOR
local INTERACTIVE_SERVER_LABEL = _G.INTERACTIVE_SERVER_LABEL

-- DB, event handler frame, Players name, Open event DB shortcut, ElvUI check
local db, loopFrame, eventFrame, playerName, openEvent, Elv, interval

-- Default settings
local DDDMItems, TDDMItems, defaultDifficulty, defaultLootTreshold = {}, {}, 0, 0
--	1	"Normal"			2	"Heroic"				3	"10 Player"
--	4	"25 Player"			5	"10 Player (Heroic)"	6	"25 Player (Heroic)"
--	7	"Looking For Raid"	8	"Challenge Mode"		9	"40 Player"
--	10	nil					11	"Heroic Scenario"		12	"Normal Scenario"
--	13	nil					14	"Flexible"											^-- MoP
--	14	"Normal Raid"		15	"Heroic Raid"			16	"Mythic Raid"			<-- WoD
for i = 14, 16 do -- Insert difficulties and their indexes to the table
	table.insert(DDDMItems, {name = GetDifficultyInfo(i), diff = i}) -- Localized
	if defaultDifficulty == 0 then -- Make the first default
		defaultDifficulty = i
	end
end
--	0 - Poor	1 - Common		2 - Uncommon	3 - Rare		<-- Gives error if you try to set treshold
--	4 - Epic	5 - Legendary	6 - Artifact	7 - Heirloom	<-- under 2, so we only take 2-5.
for i = 2, 5 do -- Insert treshold levels and their indexes to the table
	table.insert(TDDMItems, {name = _G["ITEM_QUALITY"..i.."_DESC"], tres = i}) -- Localized
	if defaultLootTreshold == 0 then -- Make the first default
		defaultLootTreshold = i
	end
end

local defaultDB = { -- This table defines the addon's default settings:
	events = {},
	config = {
		elvReskin = false,
		overlay = false,
		defaultView = true,
		defaultConfirm = true,
		defaultDifficulty = defaultDifficulty,
		defaultLootMethod = 6, -- Personal
		defaultLootTreshold = defaultLootTreshold,
		sendWhisper = true,
		InvWhisper = L.originalInvWhisper,
	},
}

local lootMethods = { -- Loot methods for _setRaidUp()
	"freeforall",
	"roundrobin",
	"master",
	"group",
	"needbeforegreed",
	"personalloot"
}
local MDDMItems = { -- Loot methods for MDDM and DMS
	strmatch(LOOT_FREE_FOR_ALL, ": ([%a%s]+)"),
	strmatch(LOOT_ROUND_ROBIN, ": ([%a%s]+)"),
	strmatch(LOOT_MASTER_LOOTER, ": ([%a%s]+)"),
	strmatch(LOOT_GROUP_LOOT, ": ([%a%s]+)"),
	strmatch(LOOT_NEED_BEFORE_GREED, ": ([%a%s]+)"),
	strmatch(LOOT_PERSONAL_LOOT, ": ([%a%s]+)")
}


-- Role table, Mass Invite table
local role, inviteTable = {}, {}

-- Create role sub-tables
local rolesList = { "Tanks", "Healers", "Melee", "Ranged", "Signup", "Standby" } -- Just to make updating all the Faux Scrollers easier
for _, v in ipairs(rolesList) do
	role[v] = {} -- Sub-tables to hold players for different roles
end

-- Titlebar
local selected = false -- Selected titlebar
local TBC = { -- Titlebar colors
	["select"] = {1, 0, 0, 0.5}, -- Color when Selected
	["deselect"] = {0.3, 0.3, 0.3, 0.75} -- Color when Deselected
	-- Rest of the colors appear only once or so and can be hard coded.
}

-- NormTime- and Round-Functions, Frames and local FrameFunctions scoped outside of the do-blocks
local _normTime, _round, Main, Think, InvBars, Cals, Tabs, Configs
local _localFunctions = {}

-- EVENTS ---------------------------------------------------------------------
-------------------------------------------------------------------------------
do -- Events block
	loopFrame = CreateFrame("Frame") -- Loop frame, using this incase Player opens different Event while loopFrame is running OnUpdate, it doesn't get overwritten.
	eventFrame = CreateFrame("Frame") -- Event frame
	eventFrame:SetScript("OnEvent", function(self, event, ...) return self[event](self, ...) end)
	eventFrame:RegisterEvent("ADDON_LOADED")
	eventFrame:RegisterEvent("PLAYER_LOGIN")

	function eventFrame:ADDON_LOADED(addon) -- Initialize
		if addon ~= ADDON_NAME then return end
		self:UnregisterEvent("ADDON_LOADED")

		if DEBUGGING then Debug("%s ADDON_LOADED", _normTime()) end -- Debug

		if DEBUGGING then Debug("%s initDB", _normTime()) end -- Debug
		LOIHCalDB = ns:initDB(LOIHCalDB, defaultDB)
		db = LOIHCalDB

		if db then
			if DEBUGGING then Debug("- DB loaded") end -- Debug
		else
			if DEBUGGING then Debug("- ??? DB not loaded") end -- Debug
		end

		if DEBUGGING then Debug("- Settings:\n-- Skin: %s\n-- Overlay: %s\n-- View: %s\n-- Confirm: %s\n-- Difficulty: %i\n-- Method: %s\n-- Treshold: %i\n-- Whisper: %s", tostring(db.config.elvReskin), tostring(db.config.overlay), tostring(db.config.defaultView), tostring(db.config.defaultConfirm), db.config.defaultDifficulty, db.config.defaultLootMethod, db.config.defaultLootTreshold, tostring(db.config.sendWhisper)) end -- Debug

		-- Initialize done, unset function
		self.ADDON_LOADED = nil
	end

	function eventFrame:PLAYER_LOGIN() -- "Enable"
		self:UnregisterEvent("PLAYER_LOGIN")

		if DEBUGGING then Debug("%s PLAYER_LOGIN", _normTime()) end -- Debug

		self:RegisterEvent("CALENDAR_EVENT_ALARM")
		self:RegisterEvent("CALENDAR_OPEN_EVENT")
		self:RegisterEvent("CALENDAR_UPDATE_EVENT")
		self:RegisterEvent("CALENDAR_UPDATE_INVITE_LIST")
		self:RegisterEvent("QUEST_LOG_UPDATE") -- Will fire after reload ui or on normal login

		Elv = IsAddOnLoaded("ElvUI")

		ns:CreateUI()
		ns.CreateUI = nil

		-- Hide Hooks
		CalendarViewEventFrame:HookScript("OnHide", function(self) -- ViewEventFrame OnHide Hook
			CalendarViewEventFrame:SetAlpha(1)
			Main:Hide()
			Tabs:Hide()
		end)
		CalendarCreateEventFrame:HookScript("OnHide", function(self) -- CreateEventFrame OnHide Hook
			CalendarCreateEventFrame:SetAlpha(1)
			Main:Hide()
			Tabs:Hide()
		end)

		playerName = UnitName("player")

		ns:PurgeDB()

		-- Enable done, unset function
		self.PLAYER_LOGIN = nil
	end

	function eventFrame:QUEST_LOG_UPDATE()
		self:UnregisterEvent("QUEST_LOG_UPDATE")

		if DEBUGGING then Debug("%s QUEST_LOG_UPDATE", _normTime()) end -- Debug

		local _, month, day, year = CalendarGetDate()
		CalendarSetAbsMonth(month, year)
		OpenCalendar() -- Requests calendar information from the server. Does not open the calendar frame.
		-- Triggers CALENDAR_UPDATE_EVENT_LIST when your query has finished processing on the server and new calendar information is available.
		self.QUEST_LOG_UPDATE = nil
	end

	function eventFrame:CALENDAR_EVENT_ALARM(...) -- 15mins to event, work on this after everything works
		local title, hour, minute = ...

		if DEBUGGING then Debug("%s CALENDAR_EVENT_ALARM \"%s\", %i, %i", _normTime(), title, hour, minute) end -- Debug
	end

	function eventFrame:CALENDAR_OPEN_EVENT(type)
		if DEBUGGING then Debug("%s CALENDAR_OPEN_EVENT \"%s\"", _normTime(), tostring(type or "!...")) end -- Debug
		
		if type ~= "PLAYER" and type ~= "GUILD_EVENT" then return end

		if CalendarViewEventFrame:IsShown() then
			Main:SetFrameStrata("HIGH")
			Main:SetFrameLevel(CalendarViewEventFrame:GetFrameLevel()+7)

			_localFunctions.renameTabs()
			Main:Show()
			Tabs:Show()
			Configs:Hide()

			if db.config.overlay then
				Main:SetPoint("TOPLEFT", CalendarViewEventFrame, "TOPLEFT")
				if db.config.defaultView then
					CalendarViewEventFrame:SetAlpha(0)

					PanelTemplates_SetTab(Tabs, 2)
					Cals:Show()
					Tabs:SetPoint("TOP", Main, "BOTTOM")
				else
					CalendarViewEventFrame:SetAlpha(1)

					PanelTemplates_SetTab(Tabs, 1)
					Main:Hide()
					Cals:Hide()
					if Elv then
						Tabs:SetPoint("TOP", CalendarViewEventFrame, "BOTTOM")
					else
						Tabs:SetPoint("TOP", CalendarViewEventFrame, "BOTTOM", 0, 4)
					end
				end
			else
				CalendarViewEventFrame:SetAlpha(1)

				Main:SetPoint("TOPLEFT", CalendarViewEventFrame, "TOPRIGHT", 26, 0)
				PanelTemplates_SetTab(Tabs, 1)
				Cals:Show()
				Tabs:SetPoint("TOP", Main, "BOTTOM")
			end
		else
			Main:SetFrameStrata("HIGH")
			Main:SetFrameLevel(CalendarCreateEventFrame:GetFrameLevel()+7)

			_localFunctions.renameTabs()
			Main:Show()
			Tabs:Show()
			Configs:Hide()

			if db.config.overlay then
				Main:SetPoint("TOPLEFT", CalendarCreateEventFrame, "TOPLEFT")
				if db.config.defaultView then
					CalendarCreateEventFrame:SetAlpha(0)

					PanelTemplates_SetTab(Tabs, 2)
					Cals:Show()
					Tabs:SetPoint("TOP", Main, "BOTTOM")
				else
					CalendarCreateEventFrame:SetAlpha(1)

					PanelTemplates_SetTab(Tabs, 1)
					Main:Hide()
					Cals:Hide()
					if Elv then
						Tabs:SetPoint("TOP", CalendarCreateEventFrame, "BOTTOM")
					else
						Tabs:SetPoint("TOP", CalendarCreateEventFrame, "BOTTOM", 0, 4)
					end
				end
			else
				CalendarCreateEventFrame:SetAlpha(1)

				Main:SetPoint("TOPLEFT", CalendarCreateEventFrame, "TOPRIGHT", 26, 0)
				PanelTemplates_SetTab(Tabs, 1)
				Cals:Show()
				Tabs:SetPoint("TOP", Main, "BOTTOM")
			end
		end

		if selected then -- Disable selected group on event change
			selected.t:SetTexture(TBC.deselect[1], TBC.deselect[2], TBC.deselect[3], TBC.deselect[4])
			selected = false 
		end

		-- Calendar might be busy waiting server for event info, let's wait...
		Think:Show()
		local timer = 0
		eventFrame:SetScript("OnUpdate", function(self, elapsed)
			timer = timer + elapsed
			local _, namesReady = CalendarEventGetNumInvites() -- This should prevent updatePlayers running before getting all the names from server
			while timer >= 0.25 do
				if not CalendarIsActionPending() and namesReady then -- Can I have moment of your time?
					eventFrame:SetScript("OnUpdate", nil)
					Think:Hide()
					ns:updatePlayers() -- Update DB and role views
				end
				timer = timer - 0.25
			end
		end)
	end

	function eventFrame:GROUP_ROSTER_UPDATE()
		if DEBUGGING then Debug("%s GROUP_ROSTER_UPDATE", _normTime()) end -- Debug

		local function _setTreshold()
			-- Calling this while the loot method is changing will revert to the original loot method.
			-- To account for this I simply set the loot threshold 2 second after calling API_SetLootMethod
			SetLootThreshold(openEvent.treshold)
		end

		if GetNumGroupMembers() > 0 and not IsInRaid() then -- In group but not in raid
			ConvertToRaid()

			-- Set up raid difficulty and loot settings
			SetRaidDifficultyID(openEvent.difficulty)
			if lootMethods[openEvent.method] == "master" then
				SetLootMethod(lootMethods[openEvent.method], "Player")
			else
				SetLootMethod(lootMethods[openEvent.method])
			end
			C_Timer.After(2, _setTreshold)
		else -- Magic is done
			self:UnregisterEvent("GROUP_ROSTER_UPDATE")
		end
	end

	function eventFrame:CALENDAR_UPDATE_EVENT()
		if DEBUGGING then Debug("%s CALENDAR_UPDATE_EVENT", _normTime()) end -- Debug

		ns:updatePlayers() -- Update DB and role views
	end

	function eventFrame:CALENDAR_UPDATE_INVITE_LIST()
		if DEBUGGING then Debug("%s CALENDAR_UPDATE_INVITE_LIST", _normTime()) end -- Debug

		ns:updatePlayers() -- Update DB and role views
	end
end -- Events block


-- DATABASE -------------------------------------------------------------------
-------------------------------------------------------------------------------
do -- DB functions
	function ns:initDB(db, defaults) -- This function copies values from one table into another:
		if type(db) ~= "table" then db = {} end
		if type(defaults) ~= "table" then return db end
		for k, v in pairs(defaults) do
			if type(v) == "table" then
				db[k] = ns:initDB(db[k], v)
			elseif type(v) ~= type(db[k]) then
				db[k] = v
			end
		end
		return db
	end

	function ns:PurgeDB() -- Remove old stuff (last year and last month) from DB
		if DEBUGGING then Debug("%s PurgeDB", _normTime()) end -- Debug

		if type(db.events) ~= "table" then return end

		local events = db.events
		local _, month, day, year = CalendarGetDate()
		if events and year and month then -- Got DB and current date info
			for k, v in pairs(events) do
				if k < year then
					if DEBUGGING then Debug("- Remove year %i, current year %i.", k, year) end -- Debug

					events[k] = nil
				else
					if DEBUGGING then Debug("- Keep year %i, current year %i.", k, year) end -- Debug

					for i, _ in pairs(v) do
						if i < month then
							if DEBUGGING then Debug("- Remove month %i, current month %i.", i, month) end -- Debug

							events[k][i] = nil
						else
							if DEBUGGING then Debug("- Keep month %i, current month %i.", i, month) end -- Debug
						end
					end	
				end
			end
		end
	end

	function ns:ResetDB() -- Nuke it, work it, make it, do it, makes us, Harder, Better, Faster, Stronger
		wipe(openEvent)
		wipe(db)
		ReloadUI()
	end
end -- DB functions


-- FUNCTIONS ------------------------------------------------------------------
-------------------------------------------------------------------------------
do -- Functions block
	function ns:elvSkin(frame, type)
		local E, L, V, P, G = unpack(ElvUI)
		local S = E:GetModule('Skins')

		if type == "frame" then
			frame:StripTextures()
			--frame:SetTemplate("Default", true)
			frame:SetTemplate("Transparent", true)
		elseif type == "button" then
			S:HandleButton(frame, true)
		elseif type == "scrollbar" then
			S:HandleScrollBar(frame)
		elseif type == "tab" then
			S:HandleTab(frame)
		elseif type == "ddm" then
			S:HandleDropDownBox(frame, frame:GetWidth())
		elseif type == "checkbox" then
			S:HandleCheckBox(frame)
		elseif type == "close" then
			S:HandleCloseButton(frame)
		elseif type == "editbox" then
			S:HandleEditBox(frame)
		end
	end

	-- Returns players index in Events invite list
	local function _getIndex(indexName)
		for i = 1, CalendarEventGetNumInvites() do
			local name = CalendarEventGetInvite(i)
			if name == indexName then
				return i
			end
		end
		return false
	end

	local numSignup = 0
	-- Update text on "Statusbar" on bottom
	local function _bottomText()
		if DEBUGGING then Debug("%s _bottomText", _normTime()) end -- Debug

		local numTanks, numHealers, numMelee, numRanged, numStandby = #role["Tanks"], #role["Healers"], #role["Melee"], #role["Ranged"], #role["Standby"] -- Count players in different roles
		
		local space = 1
		Cals.bottom.s:SetText(L.Signups.." "..numSignup..string.rep(" ", space).." "..L.Tanks.." "..numTanks..string.rep(" ", space).." "..L.Healers.." "..numHealers..string.rep(" ", space).." "..L.DPS.." "..numMelee + numRanged.." ("..numMelee.." + "..numRanged..") "..string.rep(" ", space).." "..L.Standbys.." "..numStandby)

		while Cals.bottom.s:GetStringWidth() < 309 do -- Expand the line by maxing out the space between text, and make it one step too long (Max width 318, but we want 5px marginal per side)
			space = space + 1
			Cals.bottom.s:SetText(L.Signups.." "..numSignup..string.rep(" ", space).." "..L.Tanks.." "..numTanks..string.rep(" ", space).." "..L.Healers.." "..numHealers..string.rep(" ", space).." "..L.DPS.." "..numMelee + numRanged.." ("..numMelee.." + "..numRanged..") "..string.rep(" ", space).." "..L.Standbys.." "..numStandby)
		end

		if space > 0 then -- Now the line is too long, so reduce space by 1, but don't make smaller than 0
			space = space - 1
		end

		if openEvent.difficulty and openEvent.difficulty == 16 and numSignup == 20 then -- Mythic and 20 players
			Cals.bottom.s:SetText(L.Signups.." "..GREEN_FONT_COLOR_CODE..numSignup..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Tanks.." "..HIGHLIGHT_FONT_COLOR_CODE..numTanks..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Healers.." "..HIGHLIGHT_FONT_COLOR_CODE..numHealers..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.DPS.." "..HIGHLIGHT_FONT_COLOR_CODE..numMelee + numRanged..FONT_COLOR_CODE_CLOSE.." ("..HIGHLIGHT_FONT_COLOR_CODE..numMelee..FONT_COLOR_CODE_CLOSE.." + "..HIGHLIGHT_FONT_COLOR_CODE..numRanged..FONT_COLOR_CODE_CLOSE..") "..string.rep(" ", space).." "..L.Standbys.." "..HIGHLIGHT_FONT_COLOR_CODE..numStandby..FONT_COLOR_CODE_CLOSE)
		elseif openEvent.difficulty and openEvent.difficulty == 16 then -- Mythic and not 20 players
			Cals.bottom.s:SetText(L.Signups.." "..RED_FONT_COLOR_CODE..numSignup..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Tanks.." "..HIGHLIGHT_FONT_COLOR_CODE..numTanks..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Healers.." "..HIGHLIGHT_FONT_COLOR_CODE..numHealers..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.DPS.." "..HIGHLIGHT_FONT_COLOR_CODE..numMelee + numRanged..FONT_COLOR_CODE_CLOSE.." ("..HIGHLIGHT_FONT_COLOR_CODE..numMelee..FONT_COLOR_CODE_CLOSE.." + "..HIGHLIGHT_FONT_COLOR_CODE..numRanged..FONT_COLOR_CODE_CLOSE..") "..string.rep(" ", space).." "..L.Standbys.." "..HIGHLIGHT_FONT_COLOR_CODE..numStandby..FONT_COLOR_CODE_CLOSE)
		elseif numSignup >= 10 and numSignup <= 30 then -- Between 10 and 30 players
			Cals.bottom.s:SetText(L.Signups.." "..GREEN_FONT_COLOR_CODE..numSignup..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Tanks.." "..HIGHLIGHT_FONT_COLOR_CODE..numTanks..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Healers.." "..HIGHLIGHT_FONT_COLOR_CODE..numHealers..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.DPS.." "..HIGHLIGHT_FONT_COLOR_CODE..numMelee + numRanged..FONT_COLOR_CODE_CLOSE.." ("..HIGHLIGHT_FONT_COLOR_CODE..numMelee..FONT_COLOR_CODE_CLOSE.." + "..HIGHLIGHT_FONT_COLOR_CODE..numRanged..FONT_COLOR_CODE_CLOSE..") "..string.rep(" ", space).." "..L.Standbys.." "..HIGHLIGHT_FONT_COLOR_CODE..numStandby..FONT_COLOR_CODE_CLOSE)
		else -- Not between 10 and 30 players
			Cals.bottom.s:SetText(L.Signups.." "..RED_FONT_COLOR_CODE..numSignup..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Tanks.." "..HIGHLIGHT_FONT_COLOR_CODE..numTanks..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.Healers.." "..HIGHLIGHT_FONT_COLOR_CODE..numHealers..FONT_COLOR_CODE_CLOSE..string.rep(" ", space).." "..L.DPS.." "..HIGHLIGHT_FONT_COLOR_CODE..numMelee + numRanged..FONT_COLOR_CODE_CLOSE.." ("..HIGHLIGHT_FONT_COLOR_CODE..numMelee..FONT_COLOR_CODE_CLOSE.." + "..HIGHLIGHT_FONT_COLOR_CODE..numRanged..FONT_COLOR_CODE_CLOSE..") "..string.rep(" ", space).." "..L.Standbys.." "..HIGHLIGHT_FONT_COLOR_CODE..numStandby..FONT_COLOR_CODE_CLOSE)
		end
	end

	local function calStatusSort(a, b) -- Make sure the "Green" invite statuses pop on top of the list
		local x, y = a["status"], b["status"]

		if x == 2 or x == 4 or x == 7 then
			x = x + 8
		end
		if y == 2 or y == 4 or y == 7 then
			y = y + 8
		end

		if x == y then -- If same status, then return alphabetical order
			return a["name"] < b["name"]
		else
			return x > y -- Status order (with Green statuses on top of the list)
		end
	end

	local invWhisper
	--local ERR_INVITE_PLAYER_S = string.gsub(_G.ERR_CHAT_PLAYER_NOT_FOUND_S, "%%s", "(.+)") -- Debug
	local ERR_INVITE_PLAYER_S = string.gsub(_G.ERR_INVITE_PLAYER_S, "%%s", "(.+)")
	-- Filter whisper and invite spam form Mass Invite
	local function _filterMsg(self, event, msg, author, ...)
		if DEBUGGING then Debug("%s _filterMsg, %s", _normTime(), event) end -- Debug
		
		if event == "CHAT_MSG_WHISPER_INFORM" and msg == invWhisper then
			return true
		else
			return string.match(msg, ERR_INVITE_PLAYER_S)
		end
	end

	-- Count players in event DB
	function ns:countSignups()
		if DEBUGGING then Debug("%s countSignups", _normTime()) end -- Debug

		numSignup = 0
		for k, _ in pairs(openEvent["Players"]) do -- Throw them into role slots
			if openEvent["Players"][k]["role"] == "" or openEvent["Players"][k]["role"] == nil then
				openEvent["Players"][k]["role"] = "Signup"
			end
			if openEvent["Players"][k]["status"] == 3 or openEvent["Players"][k]["status"] == 5 or openEvent["Players"][k]["status"] == 9 then -- Red or Tentative
				openEvent["Players"][k]["role"] = "Signup"
			elseif openEvent["Players"][k]["status"] == 6 then -- Standby
				openEvent["Players"][k]["role"] = "Standby"
			end

			local found = false
			for i = #role[openEvent["Players"][k]["role"]], 1, -1 do
				found = false
				if openEvent["Players"][k]["name"] == role[openEvent["Players"][k]["role"]][i]["name"] then
					found = true
					break
				end
			end

			if not found then
				table.insert(role[openEvent["Players"][k]["role"]], openEvent["Players"][k])
			end

			if openEvent["Players"][k]["status"] == 2 or openEvent["Players"][k]["status"] == 4 or openEvent["Players"][k]["status"] == 7 then -- Count signups (Accepted, Confirmed and Signed Ups)
				numSignup = numSignup + 1
			end
		end

		for _, v in ipairs(rolesList) do -- Update roles
			sort(role[v], calStatusSort)
			_localFunctions.updateScrollBar(_G[v], role[v])
		end

		_bottomText() -- Update text on "Statusbar" on bottom
	end

	-- Get Players from Calendar and DB
	function ns:updatePlayers()
		if DEBUGGING then Debug("%s updatePlayers", _normTime()) end -- Debug

		local title, _, creator, _, _, _, textureIndex, _, month, day, year, hour, minute = CalendarGetEventInfo()

		if year and year == 1999 then return end -- Don't create new event in DB everytime you start creating new event.

		if title and creator and textureIndex and month and day and year and hour and minute then
			for _, v in ipairs(rolesList) do -- Wipe Roles
				wipe(role[v])
			end

			if DEBUGGING then Debug("- %i %i %i %i%i %i-%s", year, month, day, hour, minute, textureIndex, creator) end -- Debug
			
			db.events[year] = db.events[year] or {}
			db.events[year][month] = db.events[year][month] or {}
			db.events[year][month][day] = db.events[year][month][day] or {}
			db.events[year][month][day][hour..minute] = db.events[year][month][day][hour..minute] or {}
			db.events[year][month][day][hour..minute][textureIndex.."-"..creator] = db.events[year][month][day][hour..minute][textureIndex.."-"..creator] or {}
			db.events[year][month][day][hour..minute][textureIndex.."-"..creator]["Players"] = db.events[year][month][day][hour..minute][textureIndex.."-"..creator]["Players"] or {}

			openEvent = db.events[year][month][day][hour..minute][textureIndex.."-"..creator]

			if not next(openEvent["Players"]) then -- New event, create one
				if DEBUGGING then Debug("- Created new DB entry %i-%s", textureIndex, creator) end -- Debug

				openEvent.title = title -- Title for event
				Main.title.s:SetText(ADDON_NAME.." - "..title)
				openEvent.difficulty = db.config.defaultDifficulty -- Default
				openEvent.method = db.config.defaultLootMethod -- Default
				openEvent.treshold = db.config.defaultLootTreshold -- Default
				Main.DDDM:SetSelectedID(openEvent.difficulty-13) -- List items 1-4, Difficulty items 14-16
				Main.TDDM:SetSelectedID(openEvent.treshold-1) -- List items 1-4, Treshold items 2-5
				Main.MDDM:SetSelectedID(openEvent.method)

				for i = 1, CalendarEventGetNumInvites() do -- Insert names into table
					local name, level, _, classFileName, inviteStatus, modStatus = CalendarEventGetInvite(i)

					openEvent["Players"][name] = {name = name, class = classFileName, level = level, status = inviteStatus, role = "", moderator = modStatus}
				end

				-- Now we have player info, let's see if we can enable the MIB
				if openEvent and openEvent["Players"] ~= nil and openEvent["Players"] ~= "" and openEvent["Players"][playerName] ~= nil and openEvent["Players"][playerName] ~= "" and -- Fixing the nil bug
					openEvent["Players"][playerName]["moderator"] ~= nil and openEvent["Players"][playerName]["moderator"] ~= "" and -- More nil bug fixes to be sure
					(openEvent["Players"][playerName]["moderator"] == "CREATOR" or openEvent["Players"][playerName]["moderator"] == "MODERATOR") then
					Main.MIB:Enable()
				else
					Main.MIB:Disable()
				end
			else -- Old event, update it
				if DEBUGGING then Debug("- Found DB entry %i-%s, updating", textureIndex, creator) end -- Debug

				openEvent.title = title -- Update title
				Main.title.s:SetText(ADDON_NAME.." - "..title)
				Main.DDDM:SetSelectedID(openEvent.difficulty-13) --6.0
				Main.TDDM:SetSelectedID(openEvent.treshold-1)
				Main.MDDM:SetSelectedID(openEvent.method)

				for k, _ in pairs(openEvent["Players"]) do -- Delete removed names (maybe someone ragequit or changed guild)
					local found = false
					for i = 1, CalendarEventGetNumInvites() do
						found = false
						local name = CalendarEventGetInvite(i)
						if name and name ~= "" and openEvent["Players"][k]["name"] == name then
							found = true -- Mr "name" didn't ragequit
							break
						end
					end
					if not found then -- Aha! The usual suspect, remove old name
						--table.remove(openEvent["Players"][k])
						if openEvent["Players"][k]["name"] == playerName then -- If you remove yourself from event
							if DEBUGGING then Debug("- I removed myself from event %i-%s", textureIndex, creator) end -- Debug

							openEvent = nil -- Remove the whole event
							return
						end

						if DEBUGGING then Debug("- Removed \"%s\" from event %i-%s", tostring(openEvent["Players"][k]["name"]), textureIndex, creator) end -- Debug

						openEvent["Players"][k] = nil
					end
				end

				for i = 1, CalendarEventGetNumInvites() do -- Check for new friends
					local name, level, _, classFileName, inviteStatus, modStatus = CalendarEventGetInvite(i)
					if name and name ~= "" and not openEvent["Players"][name] then -- Insert new name if found
						openEvent["Players"][name] = {name = name, class = classFileName, level = level, status = inviteStatus, role = "", moderator = modStatus}
					elseif name and name ~= "" and openEvent["Players"][name] then -- Found old friend instead, let's update his/her level, inviteStatus and modStatus
						openEvent["Players"][name]["level"] = level -- Level ups?
						openEvent["Players"][name]["status"] = inviteStatus -- Did you accept the invitation or were you confirmed since we last saw?
						openEvent["Players"][name]["moderator"] = modStatus -- Ranked up to MODERATOR of event?
					else -- This shouldn't happen
						if DEBUGGING then Debug("- ??? Found unnamed character \"%s\" from event %i-%s", tostring(name), textureIndex, creator) end -- Debug
					end
				end
				-- List is now up to date

				-- Now we have the updated player info, let's see if we can enable the Mass Invite Button
				if openEvent["Players"][playerName] and (openEvent["Players"][playerName]["moderator"] == "CREATOR" or openEvent["Players"][playerName]["moderator"] == "MODERATOR") then
					Main.MIB:Enable()
				else
					Main.MIB:Disable()
				end
			end

			self:countSignups()
		end
	end

	-- Transaction players between role-tables
	function ns:transaction(sparent, source, value, tparent, target)
		if DEBUGGING then Debug("%s transaction %s (%s) - %i - %s (%s)", _normTime(), source, tostring(sparent:GetName()), value, target, tostring(tparent:GetName())) end -- Debug

		playerName = playerName or UnitName("player")

		role[source][value]["role"] = target
		openEvent["Players"][role[source][value]["name"]]["role"] = target
		if target == "Standby" then -- Change status to Standby if moved to the group
			role[source][value]["status"] = 6
			openEvent["Players"][role[source][value]["name"]]["status"] = 6
			if _getIndex(role[source][value]["name"]) and openEvent["Players"][playerName] ~= nil and (openEvent["Players"][playerName]["moderator"] == "CREATOR" or openEvent["Players"][playerName]["moderator"] == "MODERATOR") then -- Put player on Standby, but only if you have the rights
				local index = _getIndex(role[source][value]["name"])
				CalendarEventSetStatus(index, 6) -- The real stuff
			end
		elseif target ~= "Signup" then -- Change status to Confirmed if moved to other than Signup group
			if db.config.defaultConfirm and _getIndex(role[source][value]["name"]) and openEvent["Players"][playerName] ~= nil and (openEvent["Players"][playerName]["moderator"] == "CREATOR" or openEvent["Players"][playerName]["moderator"] == "MODERATOR") then -- Confirm player, but only if you have the rights
				role[source][value]["status"] = 4
				openEvent["Players"][role[source][value]["name"]]["status"] = 4

				local index = _getIndex(role[source][value]["name"])
				CalendarEventSetStatus(index, 4) -- The real stuff
			elseif not db.config.defaultConfirm and source == "Standby" then -- Not Confirming, but moving from Stand By, set to Accepted.
				role[source][value]["status"] = 2
				openEvent["Players"][role[source][value]["name"]]["status"] = 2

				local index = _getIndex(role[source][value]["name"])
				CalendarEventSetStatus(index, 2) -- The real stuff
			end
		end

		if DEBUGGING then Debug("- Transaction Role %s / %s", role[source][value]["role"], openEvent["Players"][role[source][value]["name"]]["role"]) end -- Debug

		table.insert(role[target], role[source][value])
		table.remove(role[source], value)

		_localFunctions.updateScrollBar(tparent, role[target])
		_localFunctions.updateScrollBar(sparent, role[source])

		_bottomText() -- Update text on "Statusbar" on bottom
	end

	-- Mass Invite people to the raid
	function ns:massInvite()
		if DEBUGGING then Debug("%s massInvite", _normTime()) end -- Debug
	 
		if #inviteTable == 0 then return end

		invWhisper = ADDON_NAME..": "..string.format(db.config.InvWhisper, openEvent.title)
		playerName = playerName or UnitName("player")
		if db.config.sendWhisper then
			Debug("Whisper: %s", invWhisper)
		else
			Debug("NO WHISPER!")
		end

		eventFrame:RegisterEvent("GROUP_ROSTER_UPDATE")
		ChatFrame_AddMessageEventFilter("CHAT_MSG_SYSTEM", _filterMsg)
		ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", _filterMsg)

		local timer, interval = 0.25, 0.25
		local inRaid, invited, total = 1, 0, #inviteTable + 1 -- You are alway "already in the group" and inviteTable doesn't include you

		Main.MIB:Disable() -- Just to be safe
		InvBars:Show()

		--sort(inviteTable) -- Invite in alphabetical order

		loopFrame:SetScript("OnUpdate", function(self, elapsed)
			timer = timer + elapsed
			while timer >= interval do -- Give this enough time to do the changes
				if #inviteTable == 0 then
					if DEBUGGING then Debug("- InviteTable empty") end -- Debug

					Main.MIB:Enable()
					InvBars:Hide()
					loopFrame:SetScript("OnUpdate", nil)
					eventFrame:UnregisterEvent("GROUP_ROSTER_UPDATE")
					ChatFrame_RemoveMessageEventFilter("CHAT_MSG_SYSTEM", _filterMsg)
					ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER_INFORM", _filterMsg)

					break -- To prevent InviteUnit(nil)
				end
				
				for i = 1, GetNumGroupMembers() do
					local name = GetRaidRosterInfo(i)
					for k, _ in ipairs(inviteTable) do
						if name == inviteTable[k] then -- Already in group
							table.remove(inviteTable, k) -- Not going to invite you again
							inRaid = inRaid + 1
						end
					end
				end

				if #inviteTable > 0 then -- Still someone in the table
					if DEBUGGING then Debug("- Inviting \"%s\"", tostring(inviteTable[1])) end -- Debug

					if db.config.sendWhisper then
						SendChatMessage(invWhisper, "WHISPER", nil, inviteTable[1]) -- 1. Whisper
					end
					InviteUnit(inviteTable[1]) -- 2. Invite
					local lastInvited = table.remove(inviteTable, 1) -- 3. Remove from queue

					invited = invited + 1
					InvBars.B.s:SetText(string.format("%d/%d %s", invited, total - inRaid, lastInvited))
				end
				timer = timer - interval

				if inRaid + invited > total then
					InvBars.R:SetWidth(240 * inRaid/(inRaid + invited))
					InvBars.I:SetWidth(240 * invited/(inRaid + invited))
				else
					InvBars.R:SetWidth(240 * inRaid/total)
					InvBars.I:SetWidth(240 * invited/total)
				end

				InvBars.t:SetText(string.format(L.Estimated, #inviteTable * interval))
			end
		end)
	end
end -- Functions block


-- FRAMEFUNCTIONS -------------------------------------------------------------
-------------------------------------------------------------------------------
do -- Framefunction block
	function _normTime() -- Return time in HH:MM format instead of H:M
		local hour, minute = GetGameTime()
		if hour < 10 then
			hour = "0"..hour
		end
		if minute < 10 then
			minute = "0"..minute
		end
		return hour..":"..minute
	end

	-- Round function, copied from http://lua-users.org/wiki/SimpleRound
	function _round(num, idp)
		local mult = 10^(idp or 0)
		return math.floor(num * mult + 0.5) / mult
	end

	-- Return class-color table or colorstring
	local function _colorMe(input)
		if type(input) == "string" then -- Class
			return RAID_CLASS_COLORS[input]
		else -- Numeric
			if input == 3 or input == 5 then -- Red
				return RED_FONT_COLOR_CODE
			elseif input == 6 or input == 9 then -- Orange
				return ORANGE_FONT_COLOR_CODE
			elseif input == 2 or input == 4 or input == 7 then -- Green
				return GREEN_FONT_COLOR_CODE
			else -- (1, 8) Normal text
				return NORMAL_FONT_COLOR_CODE
			end
		end
	end

	-- Selected title
	function _localFunctions.titleOnClick(self) -- Select and highlight selected Titlebar
		if DEBUGGING then Debug("%s _titleOnClick \"%s\"", _normTime(), self:GetName()) end -- Debug

		if not selected then -- Nothing selected, selecting pressed group
			selected = self
			self.t:SetTexture(TBC.select[1], TBC.select[2], TBC.select[3], TBC.select[4])
		elseif selected ~= self then -- Deselect previous group and select pressed group
			selected.t:SetTexture(TBC.deselect[1], TBC.deselect[2], TBC.deselect[3], TBC.deselect[4])
			selected = self
			self.t:SetTexture(TBC.select[1], TBC.select[2], TBC.select[3], TBC.select[4])
		else -- Deselcting already selected group
			selected = false
			self.t:SetTexture(TBC.deselect[1], TBC.deselect[2], TBC.deselect[3], TBC.deselect[4])
		end
	end

	-- Slot Button OnClick
	function _localFunctions.slotOnClick(self)
		if DEBUGGING then Debug("%s _slotOnClick \"%s\" (%i)", _normTime(), self:GetName(), self.value) end -- Debug

		if selected and selected:GetParent() ~= self:GetParent() then
			local source = self:GetParent()
			source.name = self:GetParent():GetName()
			local target = selected:GetParent()
			target.name = selected:GetParent():GetName()
			local value = self.value

			ns:transaction(source, source.name, value, target, target.name) -- Pointer, Name, Row#, Pointer, Name
		end
	end

	-- Faux Scroller
	local calStatus = { -- Invite statuses
		[1] = L.Inv, --CALENDAR_INVITESTATUS_INVITED
		[2] = L.Acc, --CALENDAR_INVITESTATUS_ACCEPTED
		[3] = L.Dec, --CALENDAR_INVITESTATUS_DECLINED
		[4] = L.Con, --CALENDAR_INVITESTATUS_CONFIRMED
		[5] = L.Out, --CALENDAR_INVITESTATUS_OUT
		[6] = L.Sta, --CALENDAR_INVITESTATUS_STANDBY
		[7] = L.Sig, --CALENDAR_INVITESTATUS_SIGNEDUP
		[8] = L.Not, --CALENDAR_INVITESTATUS_NOT_SIGNEDUP
		[9] = L.Ten --CALENDAR_INVITESTATUS_TENTATIVE
	}
	function _localFunctions.updateScrollBar(self, tbl) -- Update Slot Buttons
		local maxValue = #tbl
		local slots = 10
		if self.name == "Tanks" or self.name == "Healers" or self.name == "Standby" then
			slots = 6
		elseif self.name == nil then -- If we update from updatePlayers()
			if self:GetName() == "Tanks" or self:GetName() == "Healers" or self:GetName() == "Standby" then
				slots = 6
			end
		end

		FauxScrollFrame_Update(self.scrollBar, maxValue, slots, 15) -- Scrollbar, max items on scrollbar, max visible items on scrollbar and height of item on scrollbar
		local offset = FauxScrollFrame_GetOffset(self.scrollBar)
		for i = 1, slots do
			local value = i + offset
			if value <= maxValue then
				local row = self.rows[i]

				-- Draw the player buttons
				row.value = value
				row.name = tbl[value]["name"]
				if row.name:match("^([^%-]+)%-(.*)$") then -- Player is from another realm
					row.name = select(1, row.name:match("^([^%-]+)%-(.*)$"))..INTERACTIVE_SERVER_LABEL -- Shorten realm name into (#)
				end
				--gsub("%-.+", "*")
				row.level = tbl[value]["level"]
				row.status = tbl[value]["status"]
				row.class = tbl[value]["class"]

				row.cc = _colorMe(row.class) or {r = 1, g = 1, b = 1, colorStr = HIGHLIGHT_FONT_COLOR_CODE}
				row.sc = _colorMe(row.status) or HIGHLIGHT_FONT_COLOR_CODE


				row.fstatus:SetText(" "..row.sc..calStatus[row.status].."|r")
				row.fname:SetText("|c"..row.cc.colorStr..row.name.."|r   ")
				row.flevel:SetText(row.level.."      ")

				row.t:SetTexture(row.cc.r, row.cc.g, row.cc.b, 0.35)
				row:Show()
				
				-- Live Version
				if row.status == 2 or row.status == 4 or row.status == 7 or row.status == 6 then -- Green people + Standby
					row:Enable()
				else -- If you really want to sign these people, change their status in calendar
					row:Disable()
					row.t:SetTexture(0.25, 0.25, 0.25, 0.25) -- Gray was too light (0.5, 0.5, 0.5)
				end

			else
				self.rows[i]:Hide()
			end
		end
	end

	-- Dropdown Menu
	function _localFunctions.DDDMOnSelect(self, arg1, arg2, checked) -- On DDM item select function
		if DEBUGGING then Debug("%s _DDDMOnSelect ID %i, Val %i, Checked \"%s\"", _normTime(), self:GetID(), self.value, tostring(checked)) end -- Debug
			
		if not checked then
			Main.DDDM:SetSelectedID(self:GetID())
			openEvent.difficulty = self.value or db.config.defaultDifficulty -- To the DBase or default (3)

			ns:countSignups() -- Update bottom text for Signup coloring
		end
	end

	function _localFunctions.TDDMOnSelect(self, arg1, arg2, checked) -- On TDM item select function
		if DEBUGGING then Debug("%s _TDDMOnSelect ID %i, Val %i, Checked \"%s\"", _normTime(), self:GetID(), self.value, tostring(checked)) end -- Debug
			
		if not checked then
			Main.TDDM:SetSelectedID(self:GetID())
			openEvent.treshold = self.value or db.config.defaultLootTreshold -- To the DBase or default (2)
		end
	end

	function _localFunctions.MDDMOnSelect(self, arg1, arg2, checked) -- On MDM item select function
		if DEBUGGING then Debug("%s _MDDMOnSelect ID %i, Val %i, Checked \"%s\"", _normTime(), self:GetID(), self.value, tostring(checked)) end -- Debug
			
		if not checked then
			Main.MDDM:SetSelectedID(self:GetID())
			openEvent.method = self.value or db.config.defaultLootMethod -- To the DBase or default (6)
		end
	end

	local info = {} -- Create only once
	function _localFunctions.DDDMInitialize() -- DDDM Initialize
		if DEBUGGING then Debug("%s _DDDMInitialize", _normTime()) end -- Debug

		--local info = UIDropDownMenu_CreateInfo() -- Source of possible taints?
		wipe(info)
		info.func = _localFunctions.DDDMOnSelect

		for i = 1, #DDDMItems do
			local name = DDDMItems[i]["name"]
			local diff = DDDMItems[i]["diff"]
			info.text = name
			info.value = diff

			if not openEvent then
				info.checked = diff == Main.DDDM.selected
			else
				info.checked = diff == openEvent.difficulty
			end

			UIDropDownMenu_AddButton(info)
		end
	end

	function _localFunctions.TDDMInitialize() -- TDDM Initialize
		if DEBUGGING then Debug("%s _TDDMInitialize", _normTime()) end -- Debug

		wipe(info)
		info.func = _localFunctions.TDDMOnSelect

		for i = 1, #TDDMItems do
			local name = TDDMItems[i]["name"]
			local tres = TDDMItems[i]["tres"]
			info.text = name
			info.value = tres
			info.colorCode = ITEM_QUALITY_COLORS[tres].hex

			if not openEvent then
				info.checked = tres == Main.TDDM.selected
			else
				info.checked = tres == openEvent.treshold
			end

			UIDropDownMenu_AddButton(info)
		end
	end

	function _localFunctions.MDDMInitialize() -- MDDM Initialize
		if DEBUGGING then Debug("%s _MDDMInitialize", _normTime()) end -- Debug

		wipe(info)
		info.func = _localFunctions.MDDMOnSelect

		for i = 1, #lootMethods do
			local name = MDDMItems[i]
			local meth = i
			info.text = name
			info.value = meth

			if not openEvent then
				info.checked = meth == Main.MDDM.selected
			else
				info.checked = meth == openEvent.method
			end

			UIDropDownMenu_AddButton(info)
		end
	end

	-- Mass Invite Button
	function _localFunctions.MIBClick() -- On click function
		if DEBUGGING then Debug("%s _MIBClick", _normTime()) end -- Debug

		-- updatePlayers() checks this, but let's check again if the button is accidently still enabled
		if openEvent["Players"][playerName]["moderator"] == "CREATOR" or openEvent["Players"][playerName]["moderator"] == "MODERATOR" then
			wipe(inviteTable)
			for k, _ in pairs(openEvent["Players"]) do
				if (openEvent["Players"][k]["status"] == 4 or openEvent["Players"][k]["status"] == 2 or openEvent["Players"][k]["status"] == 7) and openEvent["Players"][k]["role"] ~= "Signup" then -- Add only signed up people with role assigned to them
					if openEvent["Players"][k]["name"] ~= playerName then -- Don't add me, I'm already here
						table.insert(inviteTable, openEvent["Players"][k]["name"])
					end
				end
			end

			ns:massInvite() -- Send info to invite function
		end
	end

	-- Think & InvBars Frame
	local visibleCals
	function _localFunctions.getState()
		if DEBUGGING then Debug("%s _getState", _normTime()) end -- Debug

		visibleCals = Cals:IsShown()

		Cals:Hide()
	end

	function _localFunctions.setState()
		if DEBUGGING then Debug("%s _setState", _normTime()) end -- Debug

		if visibleCals then
			Cals:Show()
		end
	end

	function _localFunctions.InvCancel() -- On click function
		if DEBUGGING then Debug("%s _InvCancel", _normTime()) end -- Debug

		wipe(inviteTable)
	end

	-- Tabs Frame
	-- Tab Button OnClick
	function _localFunctions.tabOnClick(self)
		if DEBUGGING then Debug("%s _tabOnClick \"%s\" (%i)", _normTime(), self:GetName(), self:GetID()) end -- Debug
		local id = self:GetID()
		PanelTemplates_SetTab(Tabs, id)

		if db.config.overlay then
			if id == 1 then
				if CalendarViewEventFrame:IsShown() then
					CalendarViewEventFrame:SetAlpha(1)
				else
					CalendarCreateEventFrame:SetAlpha(1)
				end

				Main:Hide()
				Cals:Hide()
				Configs:Hide()
				if CalendarViewEventFrame:IsShown() then
					if Elv then
						Tabs:SetPoint("TOP", CalendarViewEventFrame, "BOTTOM")
					else
						Tabs:SetPoint("TOP", CalendarViewEventFrame, "BOTTOM", 0, 4)
					end
				else
					if Elv then
						Tabs:SetPoint("TOP", CalendarCreateEventFrame, "BOTTOM")
					else
						Tabs:SetPoint("TOP", CalendarCreateEventFrame, "BOTTOM", 0, 4)
					end
				end
			elseif id == 2 then
				if CalendarViewEventFrame:IsShown() then
					CalendarViewEventFrame:SetAlpha(0)
				else
					CalendarCreateEventFrame:SetAlpha(0)
				end

				if not Main:IsShown() then
					Main:Show()
					Tabs:SetPoint("TOP", Main, "BOTTOM")
				end
				Cals:Show()
				Configs:Hide()
				Main.title.s:SetText(ADDON_NAME.." - "..openEvent.title)
			else
				if CalendarViewEventFrame:IsShown() then
					CalendarViewEventFrame:SetAlpha(0)
				else
					CalendarCreateEventFrame:SetAlpha(0)
				end

				if not Main:IsShown() then
					Main:Show()
					Tabs:SetPoint("TOP", Main, "BOTTOM")
				end
				Cals:Hide()
				Configs:Show()
				Main.title.s:SetText(string.format("%s - %s", ADDON_NAME, L.Config))
				Configs.DDS:SetSelectedID(db.config.defaultDifficulty-13) -- Won't show text on DDM without these until you change the value
				Configs.DTS:SetSelectedID(db.config.defaultLootTreshold-1) -- DDMs ain't initialized until you open the DDM part
				Configs.DMS:SetSelectedID(db.config.defaultLootMethod)
			end
		else
			if CalendarViewEventFrame:IsShown() then
				CalendarViewEventFrame:SetAlpha(1)
			else
				CalendarCreateEventFrame:SetAlpha(1)
			end

			if id == 1 then
				Cals:Show()
				Configs:Hide()
				Main.title.s:SetText(ADDON_NAME.." - "..openEvent.title)
			else
				Cals:Hide()
				Configs:Show()
				Main.title.s:SetText(string.format("%s - %s", ADDON_NAME, L.Config))
				Configs.DDS:SetSelectedID(db.config.defaultDifficulty-13) -- Won't show text on DDM without these until you change the value
				Configs.DTS:SetSelectedID(db.config.defaultLootTreshold-1) -- DDMs ain't initialized until you open the DDM part
				Configs.DMS:SetSelectedID(db.config.defaultLootMethod)
			end
		end
	end

	-- Rename tabs and Show/Hide 3rd tab
	function _localFunctions.renameTabs()
		if db.config.overlay then
			Tabs.tab3:Show()
			Tabs.tab1:SetText(L.Default)
			Tabs.tab2:SetText(ADDON_NAME)
			Tabs.tab3:SetText(L.Config)

			PanelTemplates_SetNumTabs(Tabs, 3)
		else
			Tabs.tab3:Hide()
			Tabs.tab1:SetText(ADDON_NAME)
			Tabs.tab2:SetText(L.Config)

			PanelTemplates_SetNumTabs(Tabs, 2)
		end
	end

	-- Configs Frame
	-- Reset DB
	function _localFunctions.RDBClick() -- On click function
		if DEBUGGING then Debug("%s _RDBClick", _normTime()) end -- Debug

		if IsShiftKeyDown() then
			ns:ResetDB()
		end
	end

	-- Overlay Mode
	function _localFunctions.OLClick() -- On Click function
		if DEBUGGING then Debug("%s _OLClick \"%s\"", _normTime(), tostring(Configs.OL.button:GetChecked())) end -- Debug

		db.config.overlay = Configs.OL.button:GetChecked()
		_localFunctions.renameTabs()

		if db.config.overlay then
			if CalendarViewEventFrame:IsShown() then
				CalendarViewEventFrame:SetAlpha(0)
			else
				CalendarCreateEventFrame:SetAlpha(0)
			end

			Configs.DV:Enable()
			Main:SetPoint("TOPLEFT", CalendarViewEventFrame, "TOPLEFT")
			PanelTemplates_SetTab(Tabs, 3)
		else
			if CalendarViewEventFrame:IsShown() then
				CalendarViewEventFrame:SetAlpha(1)
			else
				CalendarCreateEventFrame:SetAlpha(1)
			end

			Configs.DV:Disable()
			Main:SetPoint("TOPLEFT", CalendarViewEventFrame, "TOPRIGHT", 26, 0)
			PanelTemplates_SetTab(Tabs, 2)
		end
	end

	-- Default View
	function _localFunctions.DVClick() -- On Click function
		if DEBUGGING then Debug("%s _DVClick \"%s\"", _normTime(), tostring(Configs.DV:GetChecked())) end -- Debug

		db.config.defaultView = Configs.DV:GetChecked()
	end

	-- Default Confirm
	function _localFunctions.DCClick() -- On Click function
		if DEBUGGING then Debug("%s _DCClick \"%s\"", _normTime(), tostring(Configs.DC:GetChecked())) end -- Debug

		db.config.defaultConfirm = Configs.DC:GetChecked()
	end

	-- Default Loot settings
	-- Dropdown Menu
	function _localFunctions.DDSOnSelect(self, arg1, arg2, checked) -- On DDS item select function
		if DEBUGGING then Debug("%s _DDSOnSelect ID %i, Val %i, Checked \"%s\"", _normTime(), self:GetID(), self.value, tostring(checked)) end -- Debug
			
		if not checked then
			Configs.DDS:SetSelectedID(self:GetID())
			db.config.defaultDifficulty = self.value -- Drop (to) the (D)Base
		end
	end

	function _localFunctions.DTSOnSelect(self, arg1, arg2, checked) -- On DTS item select function
		if DEBUGGING then Debug("%s _DTSOnSelect ID %i, Val %i, Checked \"%s\"", _normTime(), self:GetID(), self.value, tostring(checked)) end -- Debug
			
		if not checked then
			Configs.DTS:SetSelectedID(self:GetID())
			db.config.defaultLootTreshold = self.value -- Drop (to) the (D)Base
		end
	end

	function _localFunctions.DMSOnSelect(self, arg1, arg2, checked) -- On DMS item select function
		if DEBUGGING then Debug("%s _DMSOnSelect ID %i, Val %i, Checked \"%s\"", _normTime(), self:GetID(), self.value, tostring(checked)) end -- Debug
			
		if not checked then
			Configs.DMS:SetSelectedID(self:GetID())
			db.config.defaultLootMethod = self.value -- Drop (to) the (D)Base
		end
	end

	function _localFunctions.DDSInitialize() -- DDS Initialize
		if DEBUGGING then Debug("%s _DDSInitialize", _normTime()) end -- Debug

		--local info = UIDropDownMenu_CreateInfo() -- Source of possible taints?
		wipe(info)
		info.func = _localFunctions.DDSOnSelect

		for i = 1, #DDDMItems do
			local name = DDDMItems[i]["name"]
			local diff = DDDMItems[i]["diff"]
			info.text = name
			info.value = diff
			info.checked = diff == Configs.DDS.selected

			UIDropDownMenu_AddButton(info)
		end
	end

	function _localFunctions.DTSInitialize() -- DTS Initialize
		if DEBUGGING then Debug("%s _DTSInitialize", _normTime()) end -- Debug

		wipe(info)
		info.func = _localFunctions.DTSOnSelect

		for i = 1, #TDDMItems do
			local name = TDDMItems[i]["name"]
			local tres = TDDMItems[i]["tres"]
			info.text = name
			info.value = tres
			info.colorCode = ITEM_QUALITY_COLORS[tres].hex
			info.checked = tres == Configs.DTS.selected

			UIDropDownMenu_AddButton(info)
		end
	end

	function _localFunctions.DMSInitialize() -- DMS Initialize
		if DEBUGGING then Debug("%s _DMSInitialize", _normTime()) end -- Debug

		wipe(info)
		info.func = _localFunctions.DMSOnSelect

		for i = 1, #lootMethods do
			local name = MDDMItems[i]
			local meth = i
			info.text = name
			info.value = meth
			info.checked = meth == Configs.DMS.selected

			UIDropDownMenu_AddButton(info)
		end
	end

	-- Whisper settings
	function _localFunctions.WSClick() -- On Click funcion
		if DEBUGGING then Debug("%s _WSClick \"%s\"", _normTime(), tostring(Configs.WS.button:GetChecked())) end -- Debug

		db.config.sendWhisper = Configs.WS.button:GetChecked()
	end

	function _localFunctions.WEBCancel() -- On ESCPressed function
		if DEBUGGING then Debug("%s _WEBCancel", _normTime()) end -- Debug

		Configs.WEB:SetText(db.config.InvWhisper)
		Configs.WEB:ClearFocus()
	end

	function _localFunctions.WEBSClick() -- On Click function
		if DEBUGGING then Debug("%s _WEBSClick", _normTime()) end -- Debug

		db.config.InvWhisper = Configs.WEB:GetText()
		Configs.WEB:SetText(db.config.InvWhisper)
		Configs.WEB:ClearFocus()
		WEB.S:Disable()
	end

	function _localFunctions.WEBRClick() -- On Click function
		if DEBUGGING then Debug("%s _WEBRClick", _normTime()) end -- Debug

		db.config.InvWhisper = L.originalInvWhisper
		Configs.WEB:SetText(db.config.InvWhisper)
		Configs.WEB:ClearFocus()
	end

	function _localFunctions.WEBOnTextChanged(self, isUserInput) -- On Text Changed function
		if DEBUGGING then Debug("%s _WEBOnTextChanged: %s", _normTime(), tostring(isUserInput)) end -- Debug
		if isUserInput then -- Enable Save-button if user changes text
			WEB.S:Enable()
		else
			WEB.S:Disable()
		end
	end

	-- Skinning settings
	function _localFunctions.SCBClick()
		if DEBUGGING then Debug("%s _SCBClick \"%s\"", _normTime(), tostring(Configs.SCB.button:GetChecked())) end -- Debug

		db.config.elvReskin = Configs.SCB.button:GetChecked()
		ReloadUI()
	end
end -- Framefunction block


-- FACTORY --------------------------------------------------------------------
-------------------------------------------------------------------------------
do -- Factory block
	-- Background for frames
	local bg = {
		bgFile = 'Interface\\DialogFrame\\UI-DialogBox-Background',
		edgeFile = 'Interface\\ChatFrame\\ChatFrameBackground',
		tile = true, tileSize = 32, edgeSize = 1, 
		insets = { left = 0, right = 0, top = 0, bottom = 0 }
	}

	-- Frame Factory
	local invisFrames = {
		[ADDON_NAME.."_Think"] = true,
		[ADDON_NAME.."_InvBars"] = true,
		[ADDON_NAME.."_Shield"] = true,
		[ADDON_NAME.."_Tabs"] = true,
		[ADDON_NAME.."_Cals"] = true,
		[ADDON_NAME.."_Configs"] = true,
	}
	function ns:_frameFactory(type, name, parent, width, height, template) -- Creates frames and buttons
		local f = CreateFrame(type, name, parent, template)
		if width and height then
			f:SetSize(width, height)
		end
		if type == "Frame" and not invisFrames[name] then
			f:SetBackdrop(bg)
			f:SetBackdropBorderColor(0.65, 0.65, 0.65, 1)
		end

		return f
	end

	-- Title Factory
	function ns:_titleFactory(parent, offset, text) -- Creates titles for frames
		local f = CreateFrame("Button", "$parentTitle", parent)
		f:SetSize(152, 15)
		f:SetPoint("TOP", 0, offset)

		local s = f:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		s:SetPoint("CENTER")
		s:SetText(text)
		f.s = s
		if text ~= " " then -- Don't create texture for Bottombar
			local t = f:CreateTexture()
			t:SetParent(f)
			t:SetAllPoints(f)
			t:SetTexture(TBC.deselect[1], TBC.deselect[2], TBC.deselect[3], TBC.deselect[4])
			f.t = t
		end

		f:SetScript("OnClick", _localFunctions.titleOnClick)

		return f
	end

	-- Icon Factory
	function ns:_iconFactory(parent, role) -- Creates role icons
		local f = CreateFrame("Button", "$parentIcon", parent)
		f:SetSize(16, 16)
		f:EnableMouse(false) -- Click-through
		if role == "TITLE" then
			f:SetScale(8) -- 128x128
			f:SetFrameLevel(parent:GetParent():GetFrameLevel()) -- Push Logo behind other frames
			f:SetPoint("CENTER")
		elseif role == "DIFFICULTY" or role == "LOOT" or role == "INVITE" then
			f:SetPoint("RIGHT", parent.s, "LEFT", -2, 0)
		else
			f:SetPoint("RIGHT", parent.s, "LEFT", 0, 1)
		end
	
		local t = f:CreateTexture()
		t:SetParent(f)
		t:SetAllPoints(f)
		if role == "TITLE" then -- Lion Head -icon for main window title
			t:SetTexture("Interface\\Timer\\Alliance-Logo")
		elseif role == "TANK" then
			t:SetTexture("Interface\\LFGFrame\\LFGRole")
			t:SetTexCoord(0.5, 0.75, 0, 1)
		elseif role == "HEALER" then
			t:SetTexture("Interface\\LFGFrame\\LFGRole")
			t:SetTexCoord(0.75, 1, 0, 1)
		elseif role == "DPS" then
			t:SetTexture("Interface\\LFGFrame\\LFGRole")
			t:SetTexCoord(0.25, 0.5, 0, 1)
		elseif role == "SIGNUP" then
			t:SetTexture("Interface\\Minimap\\ObjectIcons")
			t:SetTexCoord(0.5, 0.625, 0.125, 0.25)
		elseif role == "STANDBY" then
			t:SetTexture("Interface\\Minimap\\ObjectIcons")
			t:SetTexCoord(0.625, 0.75, 0.5, 0.625)
		elseif role == "DIFFICULTY" then
			t:SetTexture("Interface\\Minimap\\ObjectIcons")
			t:SetTexCoord(0.625, 0.75, 0.375, 0.5)
		elseif role == "LOOT" then
			--t:SetTexture("Interface\\GroupFrame\\UI-Group-MasterLooter")
			--t:SetTexCoord(0, 1, 0, 1)
			t:SetTexture("Interface\\Minimap\\ObjectIcons")
			t:SetTexCoord(0.125, 0.25, 0.25, 0.375)
		elseif role == "INVITE" then
			t:SetTexture("Interface\\Buttons\\UI-GuildButton-MOTD-Up")
			t:SetTexCoord(0, 1, 0, 1)
		end
		f.t = t

		return f
	end

	-- Slot Factory
	function ns:_slotFactory(number, parent)
		local f = CreateFrame("Button", "$parentRow"..number, parent)
		f:SetSize(150, 15)
		f:SetNormalFontObject(GameFontNormal)
		if number == 1 then
			f:SetPoint("TOP", parent, "TOP", 0, -1) -- 1px Marginal on top of the parent
		else
			f:SetPoint("TOP", parent.rows[number-1], "BOTTOM")
		end

		local s = f:CreateFontString(nil, "OVERLAY", "GameFontNormalLeft")
		s:SetPoint("CENTER")
		f.fname = s
		
		local s = f:CreateFontString(nil, "OVERLAY", "GameFontNormalLeft")
		s:SetPoint("LEFT")
		f.fstatus = s

		local s = f:CreateFontString(nil, "OVERLAY", "GameFontNormalLeft")
		s:SetPoint("RIGHT")
		f.flevel = s

		local t = f:CreateTexture()
		t:SetParent(f)
		t:SetAllPoints(f)
		t:SetTexture(1, 1, 1, 0.25)
		f.t = t

		f:SetScript("OnClick", _localFunctions.slotOnClick)

		return f
	end

	-- Faux Scroller ScrollBar Factory
	function ns:_scrollBarFactory(parent, tbl)
		local f = CreateFrame("ScrollFrame", "$parentScrollBar", parent, "FauxScrollFrameTemplate")
		--f:SetPoint("TOPLEFT", 0, -8)
		--f:SetPoint("BOTTOMRIGHT", -30, 8)
		f:SetPoint("TOPLEFT", -7, -1)
		f:SetPoint("BOTTOMRIGHT", -23, 1)

		f:SetScript("OnVerticalScroll", function(self, offset)
			--self:SetValue(offset)
			self.offset = _round(offset / 15)
			_localFunctions.updateScrollBar(parent, tbl)
		end)

		f:SetScript("OnShow", function()
			_localFunctions.updateScrollBar(parent, tbl)
		end)

		return f
	end

	-- Tab Factory
	function ns:_tabFactory(parent, number, text)
		local f = CreateFrame("Button", "$parentTab"..number, parent, "CharacterFrameTabButtonTemplate")
		f:SetID(number)
		f:SetText(text)

		if number == 1 then
			f:SetPoint("TOPLEFT", parent, "BOTTOMLEFT", 0, 2)
		elseif number == 2 then
			f:SetPoint("LEFT", parent.tab1, "RIGHT")
		elseif number == 3 then
			f:SetPoint("LEFT", parent.tab2, "RIGHT")
		end

		f:SetScript("OnClick", _localFunctions.tabOnClick)

		return f
	end
end  -- Factory block


-- UI -------------------------------------------------------------------------
-------------------------------------------------------------------------------
do -- UI block
	function ns:CreateUI()
		-- Main Frame ------
		--------------------
		--Main = self:_frameFactory("Frame", ADDON_NAME.."_UI", UIParent, 320, 500) -- Size of ViewFrame
		Main = self:_frameFactory("Frame", ADDON_NAME.."_UI", UIParent, 320, 580) -- Size of CreateFrame
		Main:EnableMouse(true) -- Disable click through.

		Main.title = self:_titleFactory(Main, -5, ADDON_NAME)
		Main.title:SetSize(260, 15)
		Main.title:Disable() -- Don't press it

		Main.role = self:_iconFactory(Main.title, "TITLE")

		Main.close = self:_frameFactory("Button", "$parent_Close", Main, false, false, "UIPanelCloseButton")
		Main.close:SetPoint("TOPRIGHT", 4, 4)

		Main.shield = self:_frameFactory("Frame", ADDON_NAME.."_Shield", Main, 24, 580) -- Prevent pressing of the class icons on the side of the default view
		Main.shield:SetPoint("LEFT", Main, "RIGHT")
		Main.shield:SetFrameStrata("HIGH")
		Main.shield:SetFrameLevel(Main.shield:GetFrameLevel()+4)
		Main.shield:EnableMouse(true) -- Disable click through.


		-- Cals Frame ------
		--------------------
		Cals = self:_frameFactory("Frame", ADDON_NAME.."_Cals", Main, 320, 580)
		Cals:SetPoint("TOPLEFT")

		Cals.bottom = self:_titleFactory(Cals, -Cals:GetHeight()+20, " ")
		Cals.bottom:SetSize(318, 15)
		Cals.bottom:Disable() -- Don't press it

		-- Tanks frame
		Main.Tanks = self:_frameFactory("Frame", "Tanks", Cals, 152, 92)
		Main.Tanks:SetPoint("TOPLEFT", 5, -45)

		Main.Tanks.title = self:_titleFactory(Main.Tanks, 15, L.Tanks)

		Main.Tanks.role = self:_iconFactory(Main.Tanks.title, "TANK")

		Main.Tanks.scrollBar = self:_scrollBarFactory(Main.Tanks, role["Tanks"], 6) -- Faux
		Main.Tanks.rows = setmetatable({}, { __index = function(t, i)
			local row = self:_slotFactory(i, Main.Tanks)
			
			rawset(t, i, row)
			return row
		end })

		-- Healers frame
		Main.Healers = self:_frameFactory("Frame", "Healers", Cals, 152, 92)
		Main.Healers:SetPoint("LEFT", Main.Tanks, "RIGHT", 6, 0)

		Main.Healers.title = self:_titleFactory(Main.Healers, 15, L.Healers)

		Main.Healers.role = self:_iconFactory(Main.Healers.title, "HEALER")

		Main.Healers.scrollBar = self:_scrollBarFactory(Main.Healers, role["Healers"], 6) -- Faux
		Main.Healers.rows = setmetatable({}, { __index = function(t, i)
			local row = self:_slotFactory(i, Main.Healers)
			
			rawset(t, i, row)
			return row
		end })

		-- Melee frame
		Main.Melee = self:_frameFactory("Frame", "Melee", Cals, 152, 152)
		Main.Melee:SetPoint("TOP", Main.Tanks, "BOTTOM", 0, -26)

		Main.Melee.title = self:_titleFactory(Main.Melee, 15, L.Melee)

		Main.Melee.role = self:_iconFactory(Main.Melee.title, "DPS")

		Main.Melee.scrollBar = self:_scrollBarFactory(Main.Melee, role["Melee"], 10) -- Faux
		Main.Melee.rows = setmetatable({}, { __index = function(t, i)
			local row = self:_slotFactory(i, Main.Melee)
			
			rawset(t, i, row)
			return row
		end })

		-- Ranged frame
		Main.Ranged = self:_frameFactory("Frame", "Ranged", Cals, 152, 152)
		Main.Ranged:SetPoint("LEFT", Main.Melee, "RIGHT", 6, 0)

		Main.Ranged.title = self:_titleFactory(Main.Ranged, 15, L.Ranged)

		Main.Ranged.role = self:_iconFactory(Main.Ranged.title, "DPS")

		Main.Ranged.scrollBar = self:_scrollBarFactory(Main.Ranged, role["Ranged"], 10) -- Faux
		Main.Ranged.rows = setmetatable({}, { __index = function(t, i)
			local row = self:_slotFactory(i, Main.Ranged)
			
			rawset(t, i, row)
			return row
		end })

		-- Signup frame
		Main.Signup = self:_frameFactory("Frame", "Signup", Cals, 152, 152)
		Main.Signup:SetPoint("TOP", Main.Melee, "BOTTOM", 0, -26)

		Main.Signup.title = self:_titleFactory(Main.Signup, 15, L.Signups)

		Main.Signup.role = self:_iconFactory(Main.Signup.title, "SIGNUP")

		Main.Signup.scrollBar = self:_scrollBarFactory(Main.Signup, role["Signup"], 10) -- Faux
		Main.Signup.rows = setmetatable({}, { __index = function(t, i)
			local row = self:_slotFactory(i, Main.Signup)
			
			rawset(t, i, row)
			return row
		end })

		-- Standby frame
		Main.Standby = self:_frameFactory("Frame", "Standby", Cals, 152, 92)
		Main.Standby:SetPoint("TOPLEFT", Main.Signup, "TOPRIGHT", 6, 0)

		Main.Standby.title = self:_titleFactory(Main.Standby, 15, L.Standbys)

		Main.Standby.role = self:_iconFactory(Main.Standby.title, "STANDBY")

		Main.Standby.scrollBar = self:_scrollBarFactory(Main.Standby, role["Standby"], 6) -- Faux
		Main.Standby.rows = setmetatable({}, { __index = function(t, i)
			local row = self:_slotFactory(i, Main.Standby)
			
			rawset(t, i, row)
			return row
		end })

		-- Difficulty Dropdown Menu
		Main.DDDM = self:_frameFactory("Button", ADDON_NAME.."_DDDM", Cals, false, false, "UIDropDownMenuTemplate")
		Main.DDDM.selected = db.config.defaultDifficulty
		UIDropDownMenu_Initialize(Main.DDDM, _localFunctions.DDDMInitialize)
		UIDropDownMenu_SetWidth(Main.DDDM, 105)
		UIDropDownMenu_JustifyText(Main.DDDM, "CENTER")
		Main.DDDM:SetPoint("TOP", Main.Standby, "BOTTOM", 0, -26)

		Main.DDDM.title = self:_titleFactory(Main.DDDM, 15, L.RaidDifficulty)
		Main.DDDM.title:SetSize(120, 15)
		Main.DDDM.title:Disable() -- Don't press it

		Main.DDDM.role = self:_iconFactory(Main.DDDM.title, "DIFFICULTY")

		-- Quick and dirty taint fix
		function Main.DDDM:SetSelectedID(id)
			self.selected = id+13 -- List items 1-4, Difficulty items 14-16
			UIDropDownMenu_SetText(Main.DDDM, DDDMItems[id]["name"])
		end

		function Main.DDDM:GetSelectedID()
			return self.selected
		end

		-- Treshold Dropdown Menu
		Main.TDDM = self:_frameFactory("Button", ADDON_NAME.."_TDDM", Cals, false, false, "UIDropDownMenuTemplate")
		Main.TDDM.selected = db.config.defaultLootTreshold
		UIDropDownMenu_Initialize(Main.TDDM, _localFunctions.TDDMInitialize)
		UIDropDownMenu_SetWidth(Main.TDDM, 105)
		UIDropDownMenu_JustifyText(Main.TDDM, "CENTER")
		Main.TDDM:SetPoint("TOP", Main.DDDM, "BOTTOM", 0, -16)

		Main.TDDM.title = self:_titleFactory(Main.TDDM, 15, L.LootSettings)
		Main.TDDM.title:SetSize(120, 15)
		Main.TDDM.title:Disable() -- Don't press it

		Main.TDDM.role = self:_iconFactory(Main.TDDM.title, "LOOT")

		-- Quick and dirty taint fix
		function Main.TDDM:SetSelectedID(id)
			self.selected = id+1 -- List items 1-4, Treshold items 2-5
			UIDropDownMenu_SetText(Main.TDDM, ITEM_QUALITY_COLORS[TDDMItems[id]["tres"]].hex..TDDMItems[id]["name"].."|r")
		end

		function Main.TDDM:GetSelectedID()
			return self.selected
		end

		-- Method Dropdown Menu
		Main.MDDM = self:_frameFactory("Button", ADDON_NAME.."_MDDM", Cals, false, false, "UIDropDownMenuTemplate")
		Main.MDDM.selected = db.config.defaultLootMethod
		UIDropDownMenu_Initialize(Main.MDDM, _localFunctions.MDDMInitialize)
		UIDropDownMenu_SetWidth(Main.MDDM, 105)
		UIDropDownMenu_JustifyText(Main.MDDM, "CENTER")
		Main.MDDM:SetPoint("TOP", Main.TDDM, "BOTTOM", 0, 10)

		-- Quick and dirty taint fix
		function Main.MDDM:SetSelectedID(id)
			self.selected = id
			UIDropDownMenu_SetText(Main.MDDM, MDDMItems[id])
		end

		function Main.MDDM:GetSelectedID()
			return self.selected
		end

		-- Mass Invite Button
		Main.MIB = self:_frameFactory("Button", ADDON_NAME.."_MIB", Cals, 125, 28, "UIPanelButtonTemplate")
		Main.MIB:SetScript("OnClick", _localFunctions.MIBClick)
		Main.MIB:SetPoint("TOP", Main.Signup, "BOTTOM", 0, -26)
		Main.MIB:SetText(L.MassInvite)

		Main.MIB.tooltipText = L.StartMassInvite

		Main.MIB.title = self:_titleFactory(Main.MIB, 15, L.MassInvite)
		Main.MIB.title:SetSize(120, 15)
		Main.MIB.title:Disable() -- Don't press it

		Main.MIB.role = self:_iconFactory(Main.MIB.title, "INVITE")


		-- Think Frame -----
		--------------------
		Think = self:_frameFactory("Frame", ADDON_NAME.."_Think", Main, 320, 580)
		Think:SetPoint("TOPLEFT")
		Think.s = Think:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		Think.s:SetPoint("CENTER")
		Think.s:SetText(L.ServerWait)
		Think:SetScript("OnShow", _localFunctions.getState)
		Think:SetScript("OnHide", _localFunctions.setState)
		Think:Hide()


		-- InvBars Frame ---
		--------------------
		InvBars = self:_frameFactory("Frame", ADDON_NAME.."_InvBars", Main, 320, 580)
		InvBars:SetPoint("TOPLEFT")

		InvBars.B = self:_frameFactory("Frame", "InvBarsEdge", InvBars, 240, 20)
		InvBars.B:SetPoint("CENTER")
		InvBars.B.s = InvBars.B:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		InvBars.B.s:SetPoint("CENTER")

		InvBars.R = InvBars.B:CreateTexture()
		InvBars.R:SetTexture(BATTLENET_FONT_COLOR.r, BATTLENET_FONT_COLOR.g, BATTLENET_FONT_COLOR.b, 0.35)
		InvBars.R:SetSize(80, 20)
		InvBars.R:SetPoint("LEFT")

		InvBars.I = InvBars.B:CreateTexture()
		InvBars.I:SetTexture(GRAY_FONT_COLOR.r, GRAY_FONT_COLOR.g, GRAY_FONT_COLOR.b, 0.35)
		InvBars.I:SetSize(80, 20)
		InvBars.I:SetPoint("LEFT", InvBars.R, "RIGHT")

		InvBars.s = InvBars:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		InvBars.s:SetPoint("BOTTOM", InvBars.B, "TOP", 0, 10)
		InvBars.s:SetText(L.SendingInvites)

		InvBars.t = InvBars:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		InvBars.t:SetPoint("TOP", InvBars.B, "BOTTOM", 0, -10)

		InvBars.Cancel = self:_frameFactory("Button", "InvBarsCancel", InvBars, 125, 28, "UIPanelButtonTemplate")
		InvBars.Cancel:SetScript("OnClick", _localFunctions.InvCancel)
		InvBars.Cancel:SetPoint("TOP", InvBars.B, "BOTTOM", 0, -26)
		InvBars.Cancel:SetText(L.Cancel)

		InvBars:SetScript("OnShow", _localFunctions.getState)
		InvBars:SetScript("OnHide", _localFunctions.setState)
		InvBars:Hide()


		-- Tabs Frame ------
		--------------------
		Tabs = self:_frameFactory("Frame", ADDON_NAME.."_Tabs", UIParent, 320, 1)
		Tabs:SetPoint("TOP", Main, "BOTTOM")

		Tabs.tab1 = self:_tabFactory(Tabs, 1, L.Default)
		Tabs.tab2 = self:_tabFactory(Tabs, 2, ADDON_NAME)
		Tabs.tab3 = self:_tabFactory(Tabs, 3, L.Config)

		_localFunctions.renameTabs()


		-- Configs Frame ---
		--------------------
		Configs = self:_frameFactory("Frame", ADDON_NAME.."_Cals", Main, 320, 580)
		Configs:SetPoint("TOPLEFT")

		-- Reset DB
		Configs.RDB = self:_frameFactory("Frame", ADDON_NAME.."_RDB", Configs, 260, 40)
		Configs.RDB:SetPoint("TOP", Main.title, "BOTTOM", 0, -55)

		Configs.RDB.button = self:_frameFactory("Button", "RDB_Button", Configs.RDB, 125, 28, "UIPanelButtonTemplate")
		Configs.RDB.button:SetPoint("TOP", Configs.RDB, "TOP", 0, -5)
		Configs.RDB.button:SetScript("OnClick", _localFunctions.RDBClick)
		Configs.RDB.button:SetText(L.ResetDB)

		Configs.RDB.button.tooltipText = string.format(L.ShiftClick, GREEN_FONT_COLOR_CODE, FONT_COLOR_CODE_CLOSE)

		Configs.RDB.title = self:_titleFactory(Configs.RDB, 15, string.format("%s %s", L.ResetDB, string.format(L.ReloadWarning, RED_FONT_COLOR_CODE, FONT_COLOR_CODE_CLOSE)))
		Configs.RDB.title:SetSize(260, 15)
		Configs.RDB.title:Disable() -- Don't press it

		-- View Mode
		Configs.OL = self:_frameFactory("Frame", ADDON_NAME.."_OL", Configs, 260, 90)
		Configs.OL:SetPoint("TOP", Configs.RDB, "BOTTOM", 0, -30)

		Configs.OL.button = self:_frameFactory("CheckButton", "OL_Button", Configs.OL, nil, nil, "InterfaceOptionsCheckButtonTemplate")
		Configs.OL.button:SetScript("OnClick", _localFunctions.OLClick)
		Configs.OL.button:SetChecked(db.config.overlay)
		Configs.OL.button:SetPoint("TOPLEFT", Configs.OL, "TOPLEFT", 5, -5)
		_G[Configs.OL.button:GetName().."Text"]:SetText(L.AttachOnTop)

		Configs.OL.button.tooltipText = L.AttachMode
		Configs.OL.button.tooltipRequirement = string.format(L.AttachModeDesc, ADDON_NAME)

		Configs.OL.title = self:_titleFactory(Configs.OL, 15, L.AttachModeTitle)
		Configs.OL.title:SetSize(260, 15)
		Configs.OL.title:Disable() -- Don't press it

		-- Default Window
		Configs.DV = self:_frameFactory("CheckButton", "DV_Button", Configs.OL, nil, nil, "InterfaceOptionsCheckButtonTemplate")
		Configs.DV:SetScript("OnClick", _localFunctions.DVClick)
		Configs.DV:SetChecked(db.config.defaultView)
		Configs.DV:SetPoint("TOPLEFT", Configs.OL.button, "BOTTOMLEFT")
		_G[Configs.DV:GetName().."Text"]:SetText(string.format(L.OpenDefault, ADDON_NAME))

		Configs.DV.tooltipText = L.DefaultView
		Configs.DV.tooltipRequirement = string.format(L.DefaultViewDesc, ADDON_NAME)

		if Configs.OL.button:GetChecked() then
			Configs.DV:Enable()
		else
			Configs.DV:Disable()
		end

		-- Default Confirm
		Configs.DC = self:_frameFactory("CheckButton", "DC_Button", Configs.OL, nil, nil, "InterfaceOptionsCheckButtonTemplate")
		Configs.DC:SetScript("OnClick", _localFunctions.DCClick)
		Configs.DC:SetChecked(db.config.defaultConfirm)
		Configs.DC:SetPoint("TOPLEFT", Configs.DV, "BOTTOMLEFT")
		_G[Configs.DC:GetName().."Text"]:SetText(L.AutoConfirm)

		Configs.DC.tooltipText = L.AutoConfirm
		Configs.DC.tooltipRequirement = L.AutoConfirmDesc

		-- Default settings for new events
		Configs.DS = self:_frameFactory("Frame", ADDON_NAME.."_DS", Configs, 260, 100)
		Configs.DS:SetPoint("TOP", Configs.OL, "BOTTOM", 0, -30)

		Configs.DDS = self:_frameFactory("Button", "DS_DDS", Configs.DS, false, false, "UIDropDownMenuTemplate")
		Configs.DDS.selected = db.config.defaultDifficulty
		UIDropDownMenu_Initialize(Configs.DDS, _localFunctions.DDSInitialize)
		UIDropDownMenu_SetWidth(Configs.DDS, 105)
		UIDropDownMenu_JustifyText(Configs.DDS, "CENTER")
		Configs.DDS:SetPoint("TOPLEFT", Configs.DS, "TOPLEFT", 5, -5)

		Configs.DDS.title = self:_titleFactory(Configs.DS, 15, L.DefaultSettingsTitle)
		Configs.DDS.title:SetSize(260, 15)
		Configs.DDS.title:Disable() -- Don't press it

		-- Quick and dirty taint fix
		function Configs.DDS:SetSelectedID(id)
			self.selected = id+13 -- List items 1-4, Difficulty items 14-16
			UIDropDownMenu_SetText(Configs.DDS, DDDMItems[id]["name"])
		end

		function Configs.DDS:GetSelectedID()
			return self.selected
		end

		Configs.DTS = self:_frameFactory("Button", "DS_DTS", Configs.DS, false, false, "UIDropDownMenuTemplate")
		Configs.DTS.selected = db.config.defaultLootTreshold
		UIDropDownMenu_Initialize(Configs.DTS, _localFunctions.DTSInitialize)
		UIDropDownMenu_SetWidth(Configs.DTS, 105)
		UIDropDownMenu_JustifyText(Configs.DTS, "CENTER")
		Configs.DTS:SetPoint("TOPLEFT", Configs.DDS, "BOTTOMLEFT")

		-- Quick and dirty taint fix
		function Configs.DTS:SetSelectedID(id)
			self.selected = id+1 -- List items 1-4, Treshold items 2-5
			UIDropDownMenu_SetText(Configs.DTS, ITEM_QUALITY_COLORS[TDDMItems[id]["tres"]].hex..TDDMItems[id]["name"].."|r")
		end

		function Configs.DTS:GetSelectedID()
			return self.selected
		end

		Configs.DMS = self:_frameFactory("Button", "DS_DMS", Configs.DS, false, false, "UIDropDownMenuTemplate")
		Configs.DMS.selected = db.config.defaultLootMethod
		UIDropDownMenu_Initialize(Configs.DMS, _localFunctions.DMSInitialize)
		UIDropDownMenu_SetWidth(Configs.DMS, 105)
		UIDropDownMenu_JustifyText(Configs.DMS, "CENTER")
		Configs.DMS:SetPoint("TOP", Configs.DTS, "BOTTOM")

		-- Quick and dirty taint fix
		function Configs.DMS:SetSelectedID(id)
			self.selected = id
			UIDropDownMenu_SetText(Configs.DMS, MDDMItems[id])
		end

		function Configs.DMS:GetSelectedID()
			return self.selected
		end

		-- Whisper settings
		Configs.WS = self:_frameFactory("Frame", ADDON_NAME.."_WS", Configs, 260, 90)
		Configs.WS:SetPoint("TOP", Configs.DS, "BOTTOM", 0, -30)

		Configs.WS.button = self:_frameFactory("CheckButton", "WS_Button", Configs.WS, nil, nil, "InterfaceOptionsCheckButtonTemplate")
		Configs.WS.button:SetScript("OnClick", _localFunctions.WSClick)
		Configs.WS.button:SetChecked(db.config.sendWhisper)
		Configs.WS.button:SetPoint("TOPLEFT", Configs.WS, "TOPLEFT", 5, -5)
		_G[Configs.WS.button:GetName().."Text"]:SetText(L.SendWhispers)

		Configs.WS.button.tooltipText = L.SendWhispers
		Configs.WS.button.tooltipRequirement = L.SendWhispersDesc

		Configs.WS.title = self:_titleFactory(Configs.WS, 15, L.WhisperSettingsTitle)
		Configs.WS.title:SetSize(260, 15)
		Configs.WS.title:Disable() -- Don't press it

		-- Whisper editbox
		Configs.WEB = CreateFrame("EditBox", "WEB", Configs.WS, "InputBoxTemplate")
		Configs.WEB:SetSize(240, 28)
		Configs.WEB:SetText(db.config.InvWhisper)
		Configs.WEB:SetAutoFocus(false)
		Configs.WEB:SetScript("OnEnterPressed", _localFunctions.WEBSClick)
		Configs.WEB:SetScript("OnEscapePressed", _localFunctions.WEBCancel)
		Configs.WEB:SetScript("OnTextChanged", _localFunctions.WEBOnTextChanged)

		Configs.WEB:SetPoint("TOPLEFT", Configs.WS.button, "BOTTOMLEFT", 5, 5)

		-- Editbox Save / Reset -buttons
		Configs.WEB.S = self:_frameFactory("Button", "WEBS", Configs.WS, 115, 28, "UIPanelButtonTemplate")
		Configs.WEB.S:SetPoint("TOPLEFT", Configs.WEB, "BOTTOMLEFT")
		Configs.WEB.S:SetScript("OnClick", _localFunctions.WEBSClick)
		Configs.WEB.S:SetText(L.Save)

		Configs.WEB.S.tooltipText = L.SaveDesc

		Configs.WEB.R = self:_frameFactory("Button", "WEBS", Configs.WS, 115, 28, "UIPanelButtonTemplate")
		Configs.WEB.R:SetPoint("TOPRIGHT", Configs.WEB, "BOTTOMRIGHT")
		Configs.WEB.R:SetScript("OnClick", _localFunctions.WEBRClick)
		Configs.WEB.R:SetText(L.DefaultReset)

		Configs.WEB.R.tooltipText = L.DefaultResetDesc

		-- Skinning
		Configs.SCB = self:_frameFactory("Frame", ADDON_NAME.."_SCB", Configs, 260, 40)
		Configs.SCB:SetPoint("TOP", Configs.WS, "BOTTOM", 0, -30)

		Configs.SCB.button = self:_frameFactory("CheckButton", "SCB_Button", Configs.SCB, nil, nil, "InterfaceOptionsCheckButtonTemplate")
		Configs.SCB.button:SetScript("OnClick", _localFunctions.SCBClick)
		Configs.SCB.button:SetChecked(db.config.elvReskin)
		Configs.SCB.button:SetPoint("TOPLEFT", Configs.SCB, "TOPLEFT", 5, -5)
		_G[Configs.SCB.button:GetName().."Text"]:SetText(string.format(L.SkinWith, "ElvUI"))

		Configs.SCB.button.tooltipText = string.format(L.SkinWith, "ElvUI")
		Configs.SCB.button.tooltipRequirement = string.format(L.SkinWithDesc, ADDON_NAME, "ElvUI")

		if Elv then
			Configs.SCB.button:Enable()
		else
			Configs.SCB.button:Disable()
		end

		Configs.SCB.title = self:_titleFactory(Configs.SCB, 15, string.format(L.SkinningTitle, "ElvUI", string.format(L.ReloadWarning, RED_FONT_COLOR_CODE, FONT_COLOR_CODE_CLOSE)))
		Configs.SCB.title:SetSize(260, 15)
		Configs.SCB.title:Disable() -- Don't press it


		-- Skin ------------
		--------------------
		if Elv and db.config.elvReskin then -- Skin that stuff
			self:elvSkin(Main, "frame")
			self:elvSkin(Main.close, "close")
			self:elvSkin(Main.Tanks, "frame")
			self:elvSkin(Main.Healers, "frame")
			self:elvSkin(Main.Melee, "frame")
			self:elvSkin(Main.Ranged, "frame")
			self:elvSkin(Main.Signup, "frame")
			self:elvSkin(Main.Standby, "frame")
			self:elvSkin(TanksScrollBarScrollBar, "scrollbar")
			self:elvSkin(HealersScrollBarScrollBar, "scrollbar")
			self:elvSkin(MeleeScrollBarScrollBar, "scrollbar")
			self:elvSkin(RangedScrollBarScrollBar, "scrollbar")
			self:elvSkin(SignupScrollBarScrollBar, "scrollbar")
			self:elvSkin(StandbyScrollBarScrollBar, "scrollbar")
			self:elvSkin(Main.DDDM, "ddm")
			self:elvSkin(Main.TDDM, "ddm")
			self:elvSkin(Main.MDDM, "ddm")
			self:elvSkin(Main.MIB, "button")
			self:elvSkin(InvBars.Cancel, "button")

			self:elvSkin(Tabs.tab1, "tab")
			self:elvSkin(Tabs.tab2, "tab")
			self:elvSkin(Tabs.tab3, "tab")

			self:elvSkin(Configs.RDB, "frame")
			self:elvSkin(Configs.OL, "frame")
			self:elvSkin(Configs.DS, "frame")
			self:elvSkin(Configs.WS, "frame")
			self:elvSkin(Configs.SCB, "frame")
			self:elvSkin(Configs.RDB.button, "button")
			self:elvSkin(Configs.OL.button, "checkbox")
			self:elvSkin(Configs.DV, "checkbox")
			self:elvSkin(Configs.DC, "checkbox")
			self:elvSkin(Configs.DDS, "ddm")
			self:elvSkin(Configs.DTS, "ddm")
			self:elvSkin(Configs.DMS, "ddm")
			self:elvSkin(Configs.WS.button, "checkbox")
			self:elvSkin(Configs.WEB, "editbox")
			self:elvSkin(Configs.WEB.S, "button")
			self:elvSkin(Configs.WEB.R, "button")
			self:elvSkin(Configs.SCB.button, "checkbox")
		end


		-- Done ------------
		--------------------
		Configs:Hide()
		Tabs:Hide()
		Cals:Hide()
		Main:Hide()
	end
end -- UI block
-- EOF